#include <iostream.h>
#include <errno.h>
#include "MC_lib.h"

MC_header *T_REGION, *R_REGION;
imc_lkid_t lock_id[MAX_LOCK_SETS];
int MC_NUM_HOSTS = 1;
int MC_HNUM = 0;
int MC_PROC = 1;
int MC_SIZE = MBYTE;
int host_sem;
struct sembuf Ssignal, Swait, Snonblock;

int get_host_num()
{
   return MC_HNUM;
}

void * getTaddr(void *raddr)
{
      return ((void *) ((ulong) raddr-(ulong)R_REGION+(ulong)T_REGION));
}

void shmemclean()
{
   int i, status;
   int index, master;
   long glob_id;
   
   
   master = R_REGION->master;
   index = R_REGION->index;
   glob_id = R_REGION->glob_id[index];
   
   if (master) {
      for(i=0;i<MC_NUM_HOSTS*MC_PROC;i++) {
         if (R_REGION->pids[i] != (long)NO_PROC ) {
            status = R_REGION->pids[i];
            wait(&status);
         }
      }    
   }
   status = imc_asdetach(glob_id);
   if (status != IMC_SUCCESS) {
      imc_perror("imc_asdetach::", status);
      exit(status);
   }
   imc_asdealloc(glob_id);
   if (status != IMC_SUCCESS) {
      imc_perror("imc_dealloc::", status);
      exit(status);
   }
}

void spawn_slaves(int argc, char ** argv, int stval, int host_idx,
                  char *hname, long *pids)
{
   char execstr[8192]; 
   char pathname[8192];
   getwd(pathname);
   int status;
   
   for (int m=stval; m < MC_PROC; m++){
      pids[host_idx*MC_PROC+m] = NO_PROC;
      bzero(execstr, 8192);
      
      //sprintf(execstr, "/usr/bin/X11/xterm -title %s -e /bin/rsh %s \"(cd %s; %s",hosts.name[i],hosts.name[i], pathname, argv[0]);
      sprintf(execstr, "/bin/rsh %s \"(cd %s; %s",
              hname,
              pathname, argv[0]);
      for(int j=0;j<argc-1;j++) {
         strcat(execstr, " ");
         strcat(execstr, argv[j+1]);
      }
      //if (m == 0)
      //sprintf(execstr, "%s -w %d)\"", execstr, host_idx*MC_PROC+m);
      sprintf(execstr, "%s -w %d)\"", execstr, host_idx*MC_PROC+m);
      //else sprintf(execstr, "%s -w -1)\"", execstr);
      //printf("spawning %s\n", execstr);
      fflush(stdout);
      if ((pids[host_idx*MC_PROC+m]=fork()) == 0) {
         status = system(execstr);
         if (status == -1){
            perror("Exec");
            exit(errno);
         }
         exit(0);
      }
      //printf("waiting in while\n");
      //while (R_REGION->valid != host_idx*MC_PROC+m);
      //printf("out of while, %d\n", R_REGION->valid);
   }
}

void shmeminit(int master, int argc, char **argv, int host_num)
{
   imc_asid_t glob_id;
   int status;
   long pids[MC_MAXHOST*MAXPROC];
   int i, j, k, len;
   char hostname[8192];
   imc_hostinfo hosts;
   long pagesize = getpagesize();
   MC_SIZE = ((MC_SIZE/pagesize)+1)*pagesize;
   //cout << "MC_SIZE " << MC_SIZE << endl;
   if (MC_NUM_HOSTS > MC_MAXHOST) MC_NUM_HOSTS = MC_MAXHOST;
   if (MC_PROC > MAXPROC) MC_PROC = MAXPROC;
   
   gethostname(hostname, 8192);
   imc_gethosts(&hosts);

//     if (host_num == 0){
//        cout << host_num << " MC SIZE " << MC_SIZE << endl;
// //       for(i=0;i < hosts.num ;i++)
// //          cout << hosts.name[i] << endl;
//     }
   
   host_sem = semget(MC_SEM_ID, 1, IPC_CREAT);
   semctl(host_sem, 0, SETVAL, 0);
   setlinebuf(stdout);
   setlinebuf(stderr);
   
   if (MC_NUM_HOSTS > hosts.num) MC_NUM_HOSTS = hosts.num;
   MC_HNUM = host_num;
   if (master) {
      pids[0] = NO_PROC;
      spawn_slaves(argc, argv, 1, 0, hostname, pids);
      for(i=0,k=1;i < hosts.num && k < MC_NUM_HOSTS;i++) {
         if (strcmp (hostname, hosts.name[i]) != 0){
            spawn_slaves(argc, argv, 0, k, hosts.name[i], pids);
            k++;
         }
      }
   }
   
   /* Get the MC identifier */
   status = imc_asalloc(MC_CHANNEL_ID, MC_SIZE, IMC_URW, 0,&glob_id, 0);
   if (status < 0) {
      imc_perror("imc_asalloc::", status);
      exit(status);
   }
   status = imc_asattach(glob_id, IMC_TRANSMIT, IMC_SHARED, 0,
                         (caddr_t *)&T_REGION);
   if (status != IMC_SUCCESS) {
      imc_perror("imc_asattach::", status);
      exit(status);
   }
   status = imc_asattach(glob_id, IMC_RECEIVE, IMC_SHARED, 0,
                         (caddr_t *) &R_REGION);
   if (status != IMC_SUCCESS) {
      imc_perror("imc_asattach::", status);
      exit(status);
   }


   if (master){
      for(i=0;i<MC_NUM_HOSTS*MC_PROC;i++){
         //          R_REGION->pids[i] = pids[i];
         R_REGION->pids[i] = -1;
      }
      for(i=1;i<MC_NUM_HOSTS*MC_PROC;i++){
         while (R_REGION->pids[i] == -1){
            //cout << host_num << "waiting for " << i << endl << flush;
            R_REGION->valid = MC_NUM_HOSTS*MC_PROC;
            T_REGION->valid = MC_NUM_HOSTS*MC_PROC;
         }
         //cout << host_num << "done " << i << endl << flush;
      } 

      mc_zero((caddr_t) R_REGION, MC_SIZE);
      mc_zero((caddr_t) T_REGION, MC_SIZE);
      R_REGION->malloc_end = (ulong)((ulong)R_REGION + sizeof(MC_header));
      T_REGION->malloc_end = R_REGION->malloc_end;
      R_REGION->sync_cnt = MC_NUM_HOSTS*MC_PROC;
      T_REGION->sync_cnt = MC_NUM_HOSTS*MC_PROC;

      T_REGION->valid = 2*MC_NUM_HOSTS*MC_PROC;
      R_REGION->valid = 2*MC_NUM_HOSTS*MC_PROC;

      for(i=0;i<MC_NUM_HOSTS*MC_PROC;i++){
         R_REGION->pids[i] = pids[i];
      }
   }
   else{
      //if (host_num == MC_NUM_HOSTS*MC_PROC-1){
      //   R_REGION->valid = MC_NUM_HOSTS*MC_PROC-1;
      //   T_REGION->valid = MC_NUM_HOSTS*MC_PROC-1;         
      //}
      while(R_REGION->valid != MC_NUM_HOSTS*MC_PROC);
      R_REGION->pids[host_num] = host_num;
      T_REGION->pids[host_num] = host_num;
      while(R_REGION->valid != 2*MC_NUM_HOSTS*MC_PROC);
   }
   
   R_REGION->master = master;
   R_REGION->index = MC_HNUM;
   R_REGION->glob_id[MC_HNUM] = glob_id;
   T_REGION->glob_id[MC_HNUM] = glob_id;
   mc_sync();
}

void semaphore_init()
{
   host_sem = semget(MC_SEM_ID, 1, IPC_CREAT);
   if (host_sem < 0){
      perror("SEMAPHORE error");
      exit(errno);
   }
   semctl(host_sem, 0, SETVAL, 1);
   Ssignal.sem_num=0;
   Ssignal.sem_op=1;
   Swait.sem_num=0;
   Swait.sem_op=-1;
   Snonblock.sem_num=0;
   Snonblock.sem_op=-1;
   Snonblock.sem_flg = IPC_NOWAIT;
}

void semaphore_deinit()
{
   semctl(host_sem, 0, IPC_RMID, 0);
}

void locks_allocate()
{
   int status, i;
   int locks = IMC_MAXLOCKS;
   
   for (i=0; i < MAX_LOCK_SETS; i++){
      status = imc_lkalloc(MC_LOCK_ID+i, &locks, IMC_LKU|IMC_LKG|IMC_LKO, 
                           IMC_CREATOR, &lock_id[i]);
      if (locks < IMC_MAXLOCKS) printf("Error: incomplete lock set %d\n", i);
      if (status == IMC_SUCCESS){
         imc_lkrelease(lock_id[i], 0);
      }
      else if (status == IMC_EXISTS)
         imc_lkalloc(MC_LOCK_ID, &locks, IMC_LKU|IMC_LKG|IMC_LKO, 
                     0, &lock_id[i]);
      else {
         imc_perror("locks_allocate::", status);
         exit(1);
      }
   }
}


void locks_deallocate()
{
   for (int i=0; i < MAX_LOCK_SETS; i++)
      imc_lkdealloc(lock_id[i]);
}

void mc_lock_acquire(int index)
{
   //only 1 process/host can call imc_lkacquire
   semop(host_sem, &Swait, 1);
   
   int sval = index/MAX_LOCKS;
   int ival = index%MAX_LOCKS;
   imc_lkacquire(lock_id[sval], ival, 0, IMC_LOCKWAIT);
}

void mc_lock_release(int index)
{
   int sval = index/MAX_LOCKS;
   int ival = index%MAX_LOCKS;
   imc_lkrelease(lock_id[sval], ival);

   //only 1 process/host can call imc_lkacquire
   semop(host_sem, &Ssignal, 1);
}

void acq_mconly(int index)
{
   int sval = index/MAX_LOCKS;
   int ival = index%MAX_LOCKS;
   imc_lkacquire(lock_id[sval], ival, 0, IMC_LOCKWAIT);
}

void rel_mconly(int index)
{
   int sval = index/MAX_LOCKS;
   int ival = index%MAX_LOCKS;
   imc_lkrelease(lock_id[sval], ival);
}

int mc_lock_test_set(int index)
{
   int sval = index/MAX_LOCKS;
   int ival = index%MAX_LOCKS;
   if (semop(host_sem, &Snonblock, 1) == 0){
      if (imc_lkacquire(lock_id[sval], ival, 0, IMC_LOCKNOWAIT) == IMC_SUCCESS)
         return 0;
   }
   return 1;
}

void mc_copy(caddr_t src, caddr_t dest, int size)
{
    int i;
    long *s, *d;

    if (size & 0x7){
       printf("Unaglined access to mc_space\n");
       exit(1);
    }
    s = (long *)src;
    d = (long *)dest;
    size = size >> 3; /* Divide by 8 */
    for(i=0;i<size;i++)
        d[i] = s[i];
}

void mc_zero(caddr_t addr, int size)
{
    int i;
    long *s;

    if (size & 0x7){
       printf("Unaglined access to mc_space\n");
       exit(1);
    }
    s = (long *)addr;
    size = size >> 3; /* Divide by 8 */
    for(i=0;i<size;i++)
        s[i] = 0;
}

// void mc_sync()
// {
//    int pnum = get_host_num();
//    R_REGION->local_sense[pnum] = 1-R_REGION->local_sense[pnum];
//    mc_lock_acquire(0);
//    int oldcnt = R_REGION->sync_cnt;
//    R_REGION->sync_cnt -= 1;
//    T_REGION->sync_cnt = R_REGION->sync_cnt;
//    mc_lock_release(0);
//    if (oldcnt == 1){
//       R_REGION->sync_cnt = MC_NUM_HOSTS*MC_PROC;
//       T_REGION->sync_cnt = MC_NUM_HOSTS*MC_PROC;
//       R_REGION->global_sense = R_REGION->local_sense[pnum];
//       T_REGION->global_sense = R_REGION->local_sense[pnum];
//    }
//    else{
//       while(R_REGION->local_sense[pnum] != R_REGION->global_sense);
//    }
// }

// void mc_sync()
// {
//    int TOT_PROC = MC_PROC*MC_NUM_HOSTS;
//    int host_num = get_host_num();
//    int proc;

//    R_REGION->local_sense[host_num] = 1-R_REGION->local_sense[host_num];
//    T_REGION->local_sense[host_num] = 1-R_REGION->local_sense[host_num];
//    if (host_num == 0){
//       for (proc = 0; proc < TOT_PROC; proc++)
//          while (R_REGION->local_sense[proc] == R_REGION->global_sense);
//       R_REGION->global_sense = 1-R_REGION->global_sense;
//       T_REGION->global_sense = 1-R_REGION->global_sense;
//    }
//    while(R_REGION->global_sense != R_REGION->local_sense[host_num]);
// }

void mc_sync()
{
   int gbl_sense = R_REGION->global_sense;
   int TOT_PROC = MC_PROC*MC_NUM_HOSTS;
   int host_num = get_host_num();
   int proc;
   
   R_REGION->local_sense[host_num] = 1-gbl_sense;
   T_REGION->local_sense[host_num] = 1-gbl_sense;
   if (host_num == 0){
      for (proc = 0; proc < TOT_PROC; proc++)
         while (R_REGION->local_sense[proc] == gbl_sense);
      R_REGION->global_sense = 1-gbl_sense;
      T_REGION->global_sense = 1-gbl_sense;
   }
   while(R_REGION->global_sense == gbl_sense);
}

int *mc_reduce(int *cntary, int lim, int mflag)
{
   int i;

   int host_num = get_host_num();
   int TOT_PROC = MC_PROC*MC_NUM_HOSTS;

   if (MC_SIZE < lim*sizeof(int)){
      perror("Increase MC size -- mc_reduce");
      exit(errno);
   }
   
   mc_sync();
   long mc_end = R_REGION->malloc_end;
   
   int *Rshar = (int *) mc_end;
   int *Tshar = (int *) getTaddr(Rshar);

   int tmp;
   int gbl_sense = R_REGION->done_global_sense;
   
   if (host_num == 0){
      for(i=0; i < lim; i++){
         Rshar[i] = cntary[i];
         Tshar[i] = cntary[i];
      }
   }
   else{
      // else, add to the mc region
      while(R_REGION->done_sense[host_num-1] == gbl_sense);
      for(i=0; i < lim; i++){
         tmp = Rshar[i]+cntary[i];
         Rshar[i] = tmp;
         Tshar[i] = tmp;
      }
   }
   R_REGION->done_sense[host_num] = 1-gbl_sense;
   T_REGION->done_sense[host_num] = 1-gbl_sense;
   //mc_sync();
   if (host_num == 0){
      for (i = 0; i < TOT_PROC; i++)
         while (R_REGION->done_sense[i] == gbl_sense);      
      R_REGION->done_global_sense = 1-gbl_sense;
      T_REGION->done_global_sense = 1-gbl_sense;
   }
   while(R_REGION->done_global_sense == gbl_sense);
   return Rshar;

}
// int *mc_reduce(int *cntary, int lim, int mflag)
// {
//    int i;

//    int host_num = get_host_num();
//    int TOT_PROC = MC_PROC*MC_NUM_HOSTS;

//    mc_sync();
//    long mc_end = R_REGION->malloc_end;
   
//    int *Rshar = (int *) mc_end;
//    int *Tshar = (int *) getTaddr(Rshar);

//    int tmp;
//    int gbl_sense = R_REGION->done_global_sense;
   
//    if (host_num == 0){
//       for(i=0; i < lim; i++){
//          Rshar[i] = cntary[i];
//          Tshar[i] = cntary[i];
//       }
//    }
//    else{
//       // else, add to the mc region
//       while(R_REGION->done_sense[host_num-1] == gbl_sense);
//       for(i=0; i < lim; i++){
//          tmp = Rshar[i]+cntary[i];
//          Rshar[i] = tmp;
//          Tshar[i] = tmp;
//       }
//    }
//    R_REGION->done_sense[host_num] = 1-R_REGION->done_global_sense;
//    T_REGION->done_sense[host_num] = R_REGION->done_sense[host_num];
//    //mc_sync();
//    if (host_num == 0){
//       for (i = 0; i < TOT_PROC; i++)
//          while (R_REGION->done_sense[i] == gbl_sense);      
//       R_REGION->done_global_sense = 1-R_REGION->done_global_sense;
//       T_REGION->done_global_sense = R_REGION->done_global_sense;
//    }
//    while(R_REGION->done_global_sense == gbl_sense);
//    return Rshar;
// }

// int *mc_reduce(int *cntary, int lim, int mflag)
// {
//    int i;
//    int mc_res;

//    int *Rshar = (int *) R_REGION->malloc_end;
//    int *Tshar = (int *) getTaddr(Rshar);

//    int oldcnt;
//    int tmp;
   
//    mc_lock_acquire(3);

//    oldcnt = R_REGION->tmpsync;
//    R_REGION->tmpsync -= 1;
//    T_REGION->tmpsync = R_REGION->tmpsync;
   
//    //if first proc to get here, simply write to the mc region
//    if (oldcnt == MC_PROC*MC_NUM_HOSTS){
//       for(i=0; i < lim; i++){
//          Rshar[i] = cntary[i];
//          Tshar[i] = cntary[i];
//       }
//    }
//    else{ // else, add to the mc region
//       while(R_REGION->valid != oldcnt+1);
//       for(i=0; i < lim; i++){
//          tmp = Rshar[i]+cntary[i];
//          Rshar[i] = tmp;
//          Tshar[i] = tmp;
//       }
//       if (oldcnt == 1){
//          R_REGION->tmpsync = MC_PROC*MC_NUM_HOSTS;
//          T_REGION->tmpsync = MC_PROC*MC_NUM_HOSTS;
//       }
//    }
//    R_REGION->valid = oldcnt;
//    T_REGION->valid = oldcnt;
//    mc_lock_release(3);
   
//    while (R_REGION->tmpsync != MC_PROC*MC_NUM_HOSTS);
//    cout << get_host_num() << "LEAVING COUTNTTT\n";
//    return Rshar;
// }



void mc_scatter(int **pcnt, int lim)
{
   int i,j;

   int host_num = get_host_num();

   if (MC_SIZE < lim*sizeof(int)){
      perror("Increase MC size -- mc_reduce");
      exit(errno);
   }
   
   mc_sync();
   long mc_end = R_REGION->malloc_end;
   int *Rshar = (int *) mc_end+host_num*lim*sizeof(int);
   int *Tshar = (int *) getTaddr(Rshar);

   for (i=0; i < lim; i++){
      Rshar[i] = pcnt[host_num][i];
      Tshar[i] = pcnt[host_num][i];
   }

   mc_sync();
}

void mc_gather(int **pcnt, int lim)
{
   int i,j;
   
   int host_num = get_host_num();
   int TOT_PROC = MC_PROC*MC_NUM_HOSTS;
   long mc_end = R_REGION->malloc_end;   
   int *tmpary;
   for (i=0; i < TOT_PROC; i++){
      if (i != host_num){
         tmpary = (int *) mc_end+i*lim*sizeof(int);
         for (j=0; j < lim; j++){
            pcnt[i][j] = tmpary[j];
         }
      }
   }
}



