#ifndef MCLIB_H
#define MCLIB_H

#ifdef __cplusplus
extern "C" {
#endif

#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <sys/ipc.h>
#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/sem.h>
   
#include "imc.h"
   
#define MBYTE (1024*1024)
#define MC_CHANNEL_ID  137
#define MC_CHANNEL_SIZE (1*MBYTE)
#define MC_COMM_BLK (2*MBYTE)  //block of MC / proc for communicating tpose DB
   
#define MC_SEM_ID 555
#define MC_LOCK_ID     222
#define MAX_LOCK_SETS 1
#define MAX_LOCKS (MAX_LOCK_SETS*IMC_MAXLOCKS)

#define MC_MAXHOST 8
#define MAXPROC 4
/* Occupies the first sizeof(MC_header) bytes of the shared address space */
   typedef struct {
      volatile long valid;
      long malloc_end;  /* The last byte to be allocated */
      volatile long local_sense[MC_MAXHOST*MAXPROC];
      volatile long done_sense[MC_MAXHOST*MAXPROC];
      volatile long done_global_sense;
      volatile long global_sense;
      volatile long sync_cnt; /*used for cluster-wide synchronization barrier*/

      /* The following are essentially private since we never
         write in the T_REGION for them */
      long master; /* Am I the master process --- boolean private */
      long index;  /* Unique index for this host */
      long glob_id[MC_MAXHOST*MAXPROC]; /*Id of shmem segment for this host */
      long pids[MC_MAXHOST*MAXPROC]; /* Pids of slaves.  Only used by master */
   } MC_header;
   
#define NO_PROC      0xdeadcafedeadcafe
   
/* Transmit and receive region addresses */
   extern MC_header *T_REGION, *R_REGION;
   extern imc_lkid_t lock_id[MAX_LOCK_SETS];
   extern int MC_NUM_HOSTS;
   extern int MC_PROC;
   extern int MC_HNUM;
   extern int MC_SIZE;

   int get_host_num();
   void *getTaddr(void *);
   void shmemclean();
   void shmeminit(int master, int argc, char **argv, int host_num);
   void semaphore_init();
   void semaphore_deinit();
   void locks_allocate();
   void locks_deallocate();
   void mc_lock_acquire(int index);
   void mc_lock_release(int index);
   int mc_lock_test_set(int index);
   void mc_sync();
   int *mc_reduce(int *cntary, int lim, int mflag);
   void mc_copy(caddr_t src, caddr_t dest, int size);
   void mc_zero(caddr_t addr, int size);
   void mc_scatter(int **pcnt, int lim);
   void mc_gather(int **pcnt, int lim);
#ifdef __cplusplus
}
#endif


#endif /*MCLIB_H*/
