#include <math.h>
#include "pclass.h"
/*
#define Debug
*/

typedef struct ginfoo{
  double gini;  /* gini index */
  float sval;  /* split value */
  int   said;  /* split atr id */
  int   satp;  /* split atr type */
  int   sptr;  /* split pointer */
} GININFO;

double Calculate_Gini(VR *Recs,int size,int *Cabove,int *Cbelow,
		int *spointer,float *svalue) 
{
  long i,j;
  double v1,v2,k,l,v,split_above,split_below,gini,mgini;

  if(size == 0) return 1.0;

  split_above = 0.0;
  split_below = 0.0;
  v1 = 0.0;
  v2 = 0.0;
  for(j=0;j<nclass;j++) {
    k = (double)Cabove[j];
    l = (double)Cbelow[j];
    split_above += k*k;
    split_below += l*l;
    v1 += k;
    v2 += l;
  }
  v = v1+v2;
  if(v2 < EPSILON)
    if(v1 < EPSILON)
      return 1.0;
    else
      mgini = 1.0-(split_above/(v1*v));
  else
    if(v1 < EPSILON)
      mgini = 1.0-split_below/(v2*v);
    else
      mgini = 1.0-split_above/(v1*v)-split_below/(v2*v);
  *spointer = 0;

  for(i=0;i<size-1;i++) {
    j = Recs[i].cid;
    v1 -= 1.0;
    v2 += 1.0;
    k = (double)Cabove[j];
    l = (double)Cbelow[j];
    split_above += 1-2*k;
    split_below += 1+2*l;
    Cabove[j]--;
    Cbelow[j]++;
    gini = 1.0 - split_above/(v1*v) - split_below/(v2*v);
    if(gini < mgini) {
      mgini = gini;
      *spointer = i+1;
    }
  }

  j = Recs[i].cid;
  v1 -= 1.0;
  v2 += 1.0;
  k = (double)Cabove[j];
  l = (double)Cbelow[j];
  split_above += 1-2*k;
  split_below += 1+2*l;
  Cabove[j]--;
  Cbelow[j]++;
  if(v1 < EPSILON)
    gini = 1.0 - split_below/(v2*v);
  else
    gini = 1.0 - split_above/(v1*v) - split_below/(v2*v);
  if(gini < mgini) {
    mgini = gini;
    *spointer = i+1;
  }
  return mgini;
}

pQueue NewQueueNode(int natr)
{
  pQueue temp;
  int i;

  temp = (pQueue)malloc(sizeof(Queue));
  temp->CatrPointer = (int *)calloc(natr,sizeof(int));
  temp->CatrSize = (int *)calloc(natr,sizeof(int));
  temp->CatrCabove = (int **)calloc(natr,sizeof(int *));
  for(i=0;i<natr;i++)
    temp->CatrCabove[i] = (int *)calloc(nclass,sizeof(int));

  return temp;
}

void ParClassify()
{
  int *CollectedCabove, *CollectedCbelow; 
  VR *TempArray1,*TempArray2; 
  pQueue BFSqLast,BFSqFirst,LastNodeAtThisLevel,qptr,qptr1;
  pTNode Parent;
  pQueue KidL,KidR;
  int LocalNoOfClasses,GlobalNoOfClasses;
  int SplitAid,SplitPointer;
  float SplitValue,v, SplitPointValue;
  int AtrBegin,AtrSize, SplittingValAtrBegin;
  int LeftKidIndex,RightKidIndex;
  int CurrentLevel,ParentIndex;
  int LeftPointer,RightPointer;
  int ic,i,j,k,l;
  int NodesAtThisLevel,NodesAtNextLevel;
  double MinGiniIndex,GiniIndex;
  GININFO *GiniInfo;
  int PerNodeHistSize,HistSize,HistBeginNode,HistBeginAtr;
  int *Hash_Table;
/* Our Approach */

/* create first node and add it to the queue */
  TreeRoot = (pTNode)malloc(sizeof(TreeNode));

  BFSqFirst = NewQueueNode(natr);
  BFSqFirst->node = TreeRoot;
  BFSqFirst->next = NULL;
  for(ic=0;ic<natr;ic++) {
    BFSqFirst->CatrPointer[ic]=0;
    BFSqFirst->CatrSize[ic]=nrec;
    for(i=0;i<nrec;i++) {
      j = Ccid(catr,ic,i);
      BFSqFirst->CatrCabove[ic][j]++;
    }
  }

  BFSqLast = BFSqFirst;

  LastNodeAtThisLevel = BFSqLast;
  CurrentLevel = 0;

  NodesAtNextLevel = 1;

  PerNodeHistSize = natr*nclass;
  
  TempArray1 = (VR *) calloc(nrec,sizeof(VR));
  TempArray2 = (VR *) calloc(nrec,sizeof(VR));

  while(BFSqFirst) {

    /* Process current level */
    NodesAtThisLevel = NodesAtNextLevel;
    NodesAtNextLevel = 0;

    HistSize = PerNodeHistSize*NodesAtThisLevel;

    CollectedCabove = (int *)calloc(HistSize,sizeof(int));
    if(!CollectedCabove) {
      printf(" Memory Crunch @ CollectedCabove (%d)! \n",HistSize);
    }
    CollectedCbelow = (int *)calloc(HistSize,sizeof(int));
    if(!CollectedCbelow) {
      printf(" Memory Crunch @ CollectedCbelow (%d)! \n",HistSize);
    }

    GiniInfo  = (GININFO *)calloc(NodesAtThisLevel,sizeof(GININFO));
    if(!GiniInfo) {
      printf(" Memory Crunch @ GiniInfo (%d)! \n", NodesAtThisLevel*sizeof(GININFO));
    }
  
  /* Gini Compute Phase.*/

  /* collect the Cbelows of all the attributes for all the nodes */
    qptr = BFSqFirst;
    ParentIndex = 0;
    HistBeginNode = 0;
    while(qptr!=LastNodeAtThisLevel->next) {
      Parent = qptr->node;
    
   /*calculate global class distribution at this node */

      for(ic=0;ic<natr;ic++) {
        HistBeginAtr = HistBeginNode+ic*nclass;
        for(i=0;i<nclass;i++) {
          CollectedCabove[HistBeginAtr+i] = 
		qptr->CatrCabove[ic][i];
          CollectedCbelow[HistBeginAtr+i] = 0;
        }
      }
#ifdef Debug
      for(ic=0;ic<natr;ic++) 
        for(i=0;i<nclass;i++) 
           printf("Cabove %d, ", qptr->CatrCabove[ic][i]);
#endif
 
      HistBeginNode += PerNodeHistSize;
      qptr = qptr->next;

    }
    
    ParentIndex = 0;
    qptr = BFSqFirst;
    HistBeginNode = 0;
  
    while(qptr!=LastNodeAtThisLevel->next) {

    /* check for the termination criterion using atr 0's histogram. 
       simultaneously calculate Cabove for atr 0 */
       
      Parent = qptr->node;

      GlobalNoOfClasses = 0;
      for(i=0;i<nclass;i++) {
        if(CollectedCabove[HistBeginNode+i]) GlobalNoOfClasses++;
      }

      Parent->MoreThanOneClasses = 0;
      if(GlobalNoOfClasses > 1) { /* split is needed at this node */

        Parent->MoreThanOneClasses = 1;
  
      /* Find "best" split. */
	/* attribute 0 */
        AtrBegin = qptr->CatrPointer[0];

        MinGiniIndex = Calculate_Gini(&(catr[0].valsrids[AtrBegin]),
      			qptr->CatrSize[0],
			&(CollectedCabove[HistBeginNode]),
			&(CollectedCbelow[HistBeginNode]),
			&SplitPointer,&SplitValue);
        SplitAid = 0;
        for(ic=1;ic<natr;ic++) {
	  HistBeginAtr = HistBeginNode+ic*nclass;
          AtrBegin = qptr->CatrPointer[ic];
	  GiniIndex = Calculate_Gini(&(catr[ic].valsrids[AtrBegin]),
      				qptr->CatrSize[ic],
				&(CollectedCabove[HistBeginAtr]),
				&(CollectedCbelow[HistBeginAtr]),
				&k,&v);

	  if(GiniIndex < MinGiniIndex) {
	    MinGiniIndex = GiniIndex;
	    SplitPointer = k;
	    SplitAid = ic;
	    SplitValue = v;
          }
        }

        GiniInfo[ParentIndex].gini = MinGiniIndex;
        GiniInfo[ParentIndex].sval = SplitValue;
        GiniInfo[ParentIndex].said = SplitAid;
        GiniInfo[ParentIndex].sptr = SplitPointer;

        ParentIndex++;
      }

      HistBeginNode += PerNodeHistSize;
      qptr = qptr->next;
    }
    
    ParentIndex = 0;
    qptr = BFSqFirst;

    while(qptr!=LastNodeAtThisLevel->next) {
      Parent = qptr->node;
      if(Parent->MoreThanOneClasses) {
        SplitValue	= GiniInfo[ParentIndex].sval;
        SplitAid 	= GiniInfo[ParentIndex].said;
        SplitPointer    = GiniInfo[ParentIndex].sptr;

        Parent->Aid = SplitAid;
        Parent->SplitVal=SplitValue;
        AtrBegin = qptr->CatrPointer[SplitAid];
        AtrSize  = qptr->CatrSize[SplitAid];

#ifdef Debug
printf("Aid=%d, Pointer = %d, AtrBegin = %d, AtrSize = %d\n", SplitAid, SplitPointer, AtrBegin, AtrSize);
#endif
      /* SplitPointer is effectively the size of left kid 
         For the processor on which the Split lies, its the
	 same as was obtained while its local gini computation.
      */

      /* create left and right child nodes */
        Parent->left  = (pTNode)malloc(sizeof(TreeNode));
        Parent->right = (pTNode)malloc(sizeof(TreeNode));

        KidL = NewQueueNode(natr);
        KidR = NewQueueNode(natr);

        KidL->node = Parent->left;
        KidR->node = Parent->right;
        qptr->left = KidL;
        qptr->right = KidR;

      /* add left and right nodes to the queue */
        BFSqLast->next = KidL;
        BFSqLast = BFSqLast->next;
        BFSqLast->next = KidR;
        BFSqLast = BFSqLast->next;
        BFSqLast->next = NULL;

      /* partition the splitting attribute's list into those for left and 
	 right children */
      /* For continuous splitting attribute this involves just updating the
         pointers for the attribute in left and right children */

        KidL->CatrPointer[SplitAid] = AtrBegin;
        KidR->CatrPointer[SplitAid] = AtrBegin+SplitPointer;
        KidL->CatrSize[SplitAid] = SplitPointer;
        KidR->CatrSize[SplitAid] = AtrSize-SplitPointer;
        NodesAtNextLevel+=2;
       
        ParentIndex++;
      } /* End of Splitting the Parent Node */

      qptr = qptr->next;
    }
    
    qptr = BFSqFirst;
    ParentIndex = 0;
    
    while(qptr!=LastNodeAtThisLevel->next) {
      Parent = qptr->node;
      if(Parent->MoreThanOneClasses) {
        SplitAid = Parent->Aid;
        KidL = qptr->left;
        KidR = qptr->right;
        LeftKidIndex = 0;
        RightKidIndex = 1;
        Hash_Table = (int *) calloc (nrec, sizeof(int));
        for (k=0; k<nrec; k++) Hash_Table[k]=-1;
SplittingValAtrBegin = qptr->CatrPointer[SplitAid];
        l = KidL->CatrPointer[SplitAid]+KidL->CatrSize[SplitAid];
        for(i=KidL->CatrPointer[SplitAid];i<l;i++) {
        	KidL->CatrCabove[SplitAid][Ccid(catr,SplitAid,i)]++;
            Hash_Table[Crid(catr, SplitAid, i)] = LeftKidIndex;
        }

        l = KidR->CatrPointer[SplitAid]+KidR->CatrSize[SplitAid];
        for(i=KidR->CatrPointer[SplitAid];i<l;i++) {
	  KidR->CatrCabove[SplitAid][Ccid(catr,SplitAid,i)]++;
          Hash_Table[Crid(catr, SplitAid, i)] = RightKidIndex;
        }
#ifdef Debug
for (k=0; k<nrec; k++)
  printf("hash_table %d,", Hash_Table[k]);
printf("\n");       
#endif 
 
        /*process non-splitting Attributes. */    
        for (ic=0; ic<natr; ic++) {
          if (ic!=SplitAid) {
            AtrBegin = qptr->CatrPointer[ic];
            AtrSize  = qptr->CatrSize[ic];
            LeftPointer = 0;
            RightPointer = 0;
            
            for(i=AtrBegin;i<AtrBegin+AtrSize;i++) {
              if (Hash_Table[Crid(catr, ic, i)] == LeftKidIndex)
                TempArray1[LeftPointer++] = catr[ic].valsrids[i];
              else
                if (Hash_Table[Crid(catr, ic, i)] == RightKidIndex)
                  TempArray2[RightPointer++] = catr[ic].valsrids[i];
            }
#ifdef Debug
printf("ic=%d, AtrBegin=%d, AtrSize=%d, LeftPointer=%d, RightPointer=%d\n", ic, AtrBegin, AtrSize, LeftPointer, RightPointer);
#endif
            i=AtrBegin;
            for(j=0;j<LeftPointer;j++,i++) {
              catr[ic].valsrids[i] = TempArray1[j];
#ifdef Debug
printf("ic =%d, %d, %d, %d\n", ic,TempArray1[j].rid, TempArray1[j].val, TempArray1[j].cid);
#endif 
              KidL->CatrCabove[ic][Ccid(catr,ic,i)]++;
            }
            for(j=0;j<RightPointer;j++,i++) {
              catr[ic].valsrids[i] = TempArray2[j];
              KidR->CatrCabove[ic][Ccid(catr,ic,i)]++;
            }

            KidL->CatrPointer[ic] = AtrBegin;
            KidR->CatrPointer[ic] = AtrBegin+LeftPointer;
            KidL->CatrSize[ic] = LeftPointer;
            KidR->CatrSize[ic] = RightPointer;
          }
        }
        free(Hash_Table);

SplitPointValue = (float)(catr[SplitAid].valsrids[SplittingValAtrBegin + LeftPointer-1].val + catr[SplitAid].valsrids[SplittingValAtrBegin + LeftPointer].val)/2;
#ifdef Debug
printf("****Level = %d, SplitAid=%d, SplitPointValue=%f, %d, %d, LeftSize=%d, RighSize=%d\n", CurrentLevel, SplitAid,SplitPointValue, SplittingValAtrBegin+LeftPointer-1, SplittingValAtrBegin + LeftPointer, LeftPointer, RightPointer); 
#endif
      }
      qptr = qptr->next;
    } /* End of the CurrentLevel processing for Splitting Attributes*/
    
    /*delete all the nodes at CurrentLevel from the queue. 
      Adjust BFSqFirst and LastNodeAtThisLevel. */ 

    qptr = BFSqFirst;
    qptr1 = BFSqFirst->next;
    while(qptr1 != LastNodeAtThisLevel->next) {
      for(i=0;i<natr;i++)
        free(qptr->CatrCabove[i]);
      free(qptr->CatrCabove);
      free(qptr->CatrSize);
      free(qptr->CatrPointer);
      free(qptr);
      qptr = qptr1;
      qptr1 = qptr->next;
    }

    BFSqFirst = LastNodeAtThisLevel->next;
    LastNodeAtThisLevel=BFSqLast;
    CurrentLevel++;

    free(CollectedCabove);
    free(CollectedCbelow);
    free(GiniInfo);

    printf("Number of Nodes at Next level = %d\n",NodesAtNextLevel);
  }

  printf("records out of %d recs\n",nrec);
  printf("******************************************************\n");
  printf("Number of Levels     = %d\n",CurrentLevel);
  printf("******************************************************\n");

}

